#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cmath>

///////////////////////////////////////
//
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// MS 977
// 345 MIDDLEFIELD RD
// MENLO PARK, CALIFORNIA 94025
// goldman@usgs.gov
//
///////////////////////////////////////

main(int argc, char** argv) {
 
   float ibm2ieee(const float ibm);
   
   double ibm2ieee(const double ibm);

   if (argc != 3) {
      printf("\n");
      printf("%s converts a standard (IBM floating point) SEG-Y file\n",argv[0]);
      printf("to a non-standard (IEEE floating point) SEGY-Y file\n");
      printf("\n");
      printf("format:  %s  input_segy_filename  output_segy_filename\n",argv[0]);
      printf("\n");
      exit(1);
   }

   FILE* in;
   FILE* out;

   char dummy[3601];  // dummy buffer for skipping unused header information

   short numsamp;
   short samprate;
   int   ntraces;

/*
short itest;
short* ptest = &itest;
unsigned char* cptr;
unsigned char tmp;
*/

   const char* input = argv[1];
   const char* output = argv[2];

   // OPEN INPUT AND OUTPUT FILES
   if ((in = fopen(input,"r")) == NULL) {
      printf("\nERROR:  cannot find file %s\n\n",input);
      exit(1);
   }
   if ((out = fopen(output,"w")) == NULL) {
      printf("\nERROR:  cannot open file %s\n\n",output);
      exit(1);
   }

   // GET NUMBER_SAMPLES, SAMPLE_RATE FROM HEADER INFORMATION
   // (SKIP TO CORRECT LOCATION IN HEADER FILE BY
   // READING IN UNUSED BYTES TO A DUMMY VARIABLE)
   fread (&dummy,    3216, 1, in);
   fwrite(&dummy,    3216, 1, out);

/*
fread (ptest,    2, 1, in);
printf("sample rate is %d\n", itest);
cptr = (unsigned char*)ptest;
tmp = cptr[0];
cptr[0] = cptr[1];
cptr[1] = tmp;
printf("sample rate is %d\n", itest);
samprate = itest;
*/

   fread (&samprate,    2, 1, in);
   fwrite(&samprate,    2, 1, out);

   fread (&dummy,       2, 1, in);
   fwrite(&dummy,       2, 1, out);

   fread (&numsamp,     2, 1, in);
   fwrite(&numsamp,     2, 1, out);

/*
fread (ptest,     2, 1, in);
printf("num samples is %d\n", itest);
cptr = (unsigned char*)ptest;
tmp = cptr[0];
cptr[0] = cptr[1];
cptr[1] = tmp;
printf("sample rate is %d\n", itest);
numsamp = itest;
*/

   fread (&dummy,     378, 1, in);
   fwrite(&dummy,     378, 1, out);

   double fsamprate = samprate / 1000.0;

   // READ TRACES
   float value;
   ntraces = 0;

   while (fread(&dummy, 240, 1, in) == 1) {

      fwrite(&dummy, 240, 1, out);

      ntraces++;

      for (int i = 0 ; i < numsamp; i++) {

         if (fread(&value, 4, 1, in) != 1) {
            printf("\nERROR:  unexpected end of segy file %s\n\n",input);
            exit(1);
         }

         value = ibm2ieee(value);

         fwrite(&value, 4, 1, out);

      }

   }

//printf("NEW %6d  %6d  %8.4f\n", ntraces, i, value);

/*
float* pvalue = &value;
cptr = (unsigned char*)pvalue;
tmp = cptr[0];
cptr[0] = cptr[3];
cptr[3] = tmp;
tmp = cptr[1];
cptr[1] = cptr[2];
cptr[2] = tmp;
printf("NEW NEW %d  %8.2f\n", i, value);
*/

   fclose(in);
   fclose(out);

   // PRINT INFO
   printf("\n");
   printf("input file %s has IBM floating point format\n",input);
   printf("\n");
   printf("input file has %d traces, each with %d samples, sample rate of %f ms\n",
      ntraces, numsamp, fsamprate);
   printf("\n");
   printf("new output file %s has IEEE floating point format\n", output);
   printf("\n");

}

float ibm2ieee(const float ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      float f;
      unsigned char c[4];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long IntMantissa =
        ((long)src.c[1] << 16)
      + ((long)src.c[2] << 8)
      + src.c[3];

   float  Mantissa  = float(IntMantissa) / float(0x1000000);
   float  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

double ibm2ieee(const double ibm) {

   // DEFINE AND SETUP UNION VARIABLES
   union value {
      double f;
      unsigned char c[8];
   };
   value src;

   // ASSIGN PARAMETER TO UNION
   src.f = ibm;

   // CONVERT TO FLOAT
   long long IntMantissa =
        ((long long)src.c[1] << 48)
      + ((long long)src.c[2] << 40)
      + ((long long)src.c[3] << 32)
      + ((long long)src.c[4] << 24)
      + ((long long)src.c[5] << 16)
      + ((long long)src.c[6] << 8)
      + src.c[7];

   double  Mantissa  = double(IntMantissa) / double(0x1000000);
   double  PosResult = Mantissa * pow( 16.0, double((src.c[0] & 0x7F) - 64) );

   if (src.c[0] & 0x80)
      return -PosResult;
   else
      return PosResult;

}

